
### PV Interface for {{ device }} ###
{% for record in records %}
record("*", "{{ record.pv }}") {
    info(Q:group, {
        "{{ pv_prefix }}PVI": {
            "pvi.{{ record.name }}.{{ record.access }}": {
                "+channel": "NAME",
                "+type": "plain",
            }
        }
    })
}
{% endfor %}
### End of PV Interface for {{ device }} ###
